from typing import List

from pandas import DataFrame


def _gen_vertical_table(df, caption, header, line_names, index, index_start):
    col_count = len(df.columns)

    cols = ''.join(['c|' for i in range(col_count)])

    # -------------begin-------------
    index_col = ''
    if index:
        index_col = 'c|'

    header_col = ''
    if header:
        header_col = f'\t\t\t\\hline\n' \
                     f'\t\t\t\\multicolumn{{{col_count}}}{{|c|}}{{{header}}}\\\\\n'

    table_begin = f'\\begin{{table}}[H]\n' \
                  f'\t\\begin{{center}}\n' \
                  f'\t\t\\begin{{tabular}}{{|{index_col}{cols}}}\n' \
                  f'{header_col}' \
                  f'\t\t\t\\hline\n'

    # -------------body-------------
    if line_names:
        body = ' & '.join(line_names)

    else:
        body = ' & '.join(df.columns)

    body += '\\\\\n\t\t\t\\hline\n'

    if index:
        body = '\t\t\tN & ' + body
    else:
        body = '\t\t\t' + body

    if index:
        for num, i in enumerate(df.values):
            body += f'\t\t\t{num + index_start} & ' + ' & '.join([str(elem) for elem in i]) + '\\\\\n\t\t\t\\hline\n'

    else:
        for num, i in enumerate(df.values):
            body += '\t\t\t' + ' & '.join([str(elem) for elem in i]) + '\\\\\n\t\t\t\\hline\n'

    # -------------caption-------------
    caption_text = ''
    if caption is not None:
        caption_text = f'\t\\caption{{{caption}}}\n'

    # -------------end-------------
    table_end = f'\t\t\\end{{tabular}}\n' \
                f'\t\\end{{center}}\n' \
                f'{caption_text}' \
                f'\\end{{table}}'

    return f'{table_begin}{body}{table_end}'


def _gen_horizontal_table(df, caption, header, line_names, index, index_start):
    # -------------begin-------------
    cols = len(df.index) + 1  # +1, так как сбоку заголовок

    header_col = ''
    if header:
        header_col = f'\t\t\t\\hline\n' \
                     f'\t\t\t\\multicolumn{{{cols}}}{{|c|}}{{{header}}}\\\\\n'

    table_begin = f'\\begin{{table}}[H]\n' \
                  f'\t\\begin{{center}}\n' \
                  f'\t\t\\begin{{tabular}}{{|{"c|" * cols}}}\n' \
                  f'{header_col}' \
                  f'\t\t\t\\hline\n'

    # -------------body-------------

    body = ""

    if index:
        body += '\t\t\t& ' + ' & '.join(
            [str(i) for i in range(index_start, cols + index_start - 1)]) + ' \\\\\n\t\t\t\\hline\n'

    for num, row in enumerate(df):
        line_names_element = row

        if line_names:
            line_names_element = line_names[num]

        body += f'\t\t\t{line_names_element} & ' + ' & '.join([str(i) for i in df[row]]) + ' \\\\\n\t\t\t\\hline\n'

    # -------------caption-------------
    caption_text = ''
    if caption is not None:
        caption_text = f'\t\\caption{{{caption}}}\n'

    # -------------end-------------
    table_end = f'\t\t\\end{{tabular}}\n' \
                f'\t\\end{{center}}\n' \
                f'{caption_text}' \
                f'\\end{{table}}'

    return f'{table_begin}{body}{table_end}'


def gen_from_dataframe(df: DataFrame,
                       caption: str = None, header: List[str] = None, line_names: List[str] = None,
                       index: bool = False, index_start: int = 1, horizontal: bool = False) -> str:
    """
    Генерирует таблицу latex из DataFame.
    Нужно лишь дописать свой заголовок таблицы (хотя можно оставить сгенерированный)

    По умолчанию в заголовке находятся имена столбцов из pandas.

    :param df: dataframe из pandas
    :param caption: название таблицы
    :param header: заголовок таблицы, располагается сверху всех строк и размером со все столбцы
    :param line_names: названий столбцов/строк в таблице
    :param index: нужна ли индексация
    :param index_start: начала индексации, по умолчанию 1
    :param horizontal: нужно ли строить горизонтальную таблицу

    :return: строку, содержащую таблицу latex, которую можно вставить в редактор
    """
    if horizontal:
        return _gen_horizontal_table(df, caption, header, line_names, index, index_start)

    else:
        return _gen_vertical_table(df, caption, header, line_names, index, index_start)



"""
Модуль отвечает за всё, что непосредственно относиться к рисованию графиков
"""
from .plotter import (
    NothingDrawError,
    pretty_plot, pretty_plot_many, show,
    savefig,
)

__all__ = ['NothingDrawError', 'pretty_plot', 'pretty_plot_many', 'show', 'savefig']



def represent_value(name, val=None, sigma=None, eps=None, units='', n=-1):
    r"""
    Форматирует значение величины для лабы в формат

    Если есть погрешности: $ name = val \\pm sigma [units], \\; \\varepsilon_{name} = eps \\% $

    Иначе: $ name = val \\pm [units] $

    :param name: Название величины
    :param val: Значение величины
    :param sigma: Абсолютная величины
    :param eps: Относительная величины
    :param units: единицы измерения
    :param n: число знаков округления, если -1, то округления нет
    :return: представление величины в формате latex для копипаста в текст лабы
    """
    # Проверка, что не все значения None
    if val is None and sigma is None and eps is None:
        raise ValueError(f'У величины {name} не переданы никакие значения')

    # Проверка, что если все не None, то выполняется тождество  val * eps == sigma
    if val is not None and sigma is not None and eps is not None and val * eps != sigma:
        raise ValueError(f'У величины {name} переданы некорректные значения')

    if units:
        units = f' \\; {units}'

    # Проверка, что есть пара не None
    if sigma is None and eps is None:
        print(f'WARNING!!! У величины {name} передано только значение.')
        return f'$ {name} = {round(val, n) if n != -1 else n}{units}$'

    if val is None and eps is None:
        raise ValueError(f'У величины {name} передана, только абсолютная погрешность')

    if val is None and sigma is None:
        raise ValueError(f'У величины {name} передана, только относительная погрешность')

    # Далее форматирование
    if eps is None and val != 0:
        eps = abs(sigma / val)

    if sigma is None:
        sigma = abs(val * eps)

    if val is None:
        if eps == 0:
            if sigma != 0:
                raise ValueError(f'У величины {name} переданы только погрешности, '
                                 f'при чем относительная равна нулю, а абсолютная нет')
            else:
                raise ValueError(f'У величины {name} переданы только погрешности, при чем относительная и '
                                 f'абсолютная равна нулю одновременно, получается неопределенность')
        else:
            print(f'WARNING!!! У величины {name} переданы только погрености, но не сама величина. '
                  f'Будет вычислено с точностью до знака Это странно.')
            val = sigma / eps

    if n != -1:
        val_ = round(val, n)
        sigma_ = round(sigma, n)
        eps_ = round(eps * 100, n)

    else:
        val_ = val
        sigma_ = sigma
        eps_ = eps * 100

    return f'$ {name} = {val_} \\pm {sigma_}{units}, \\; \\varepsilon_{{{name}}} = {eps_} \\% $'



from .base import (
    Polynomial,
    Functional,
)
from .experemental import (
    Lowess,
    Fourier,
)
from .special import (
    Linear,
    Exponential,
    Logarithmic,
)

__all__ = ['Polynomial', 'Functional', 'Linear', 'Exponential', 'Logarithmic', 'Lowess', 'Fourier']



"""
Модуль содержаший базовые апроксиматоры
"""

from warnings import warn

import numpy as np
import scipy.fftpack
import scipy.optimize

from ..utils import round_to_n, format_monoid


class Approximator:
    """
    Базовый класс апроксиматор.

    Классы нужны, чтобы сохранять разные данные, во время аппроксимации. Например коэффициенты, ошибки.
    Возможно когда-нибудь будет сделан класс для результата аппроксимации

    Для некоторых позволяют генерировать формулу в латехе
    """

    def __init__(self, points=100, left_offset=5, right_offset=5):
        """
        :param points: количество точек, которые будут на выходе
        :param left_offset: отступ от левой гриницы диапозона
        :param right_offset: отступ от правой гриницы диапозона
        """
        self.left_offset = left_offset / 100
        self.right_offset = right_offset / 100
        self.points = points
        self.meta = []

        # Данные
        self._x = np.array([])
        self._y = np.array([])
        self._xerr = np.array([])
        self._yerr = np.array([])

        # Коэффициенты апроксимации
        self.koefs = np.array([])
        # Стандартное отклонение параметров
        self.sigmas = np.array([])

        # Наиболее правдоподобная функция
        self._function = None

    def __init_subclass__(cls, **kwargs):
        """
        Костыльное добавление документации к классам наследникам
        """
        super().__init_subclass__(**kwargs)

        if not cls.__init__.__doc__:
            return

        if not cls.__doc__:
            cls.__doc__ = cls.__init__.__doc__
            return

        # для совместимости со старыми питонами
        def remove_pref(s: str, pref: str):
            if s.startswith(pref):
                return s[len(pref):]

            return s

        cls.__doc__ = '{}\n{}'.format(
            cls.__doc__,
            '\n'.join(
                [
                    remove_pref(remove_pref(line, '\t'), ' ' * 4)
                    for line in cls.__init__.__doc__.split('\n')
                ]
            )
        )

    def _gen_x_axis_with_offset(self, start, end):
        """
        Генерирует набор точек по оси абсцисс в заданом диапозоне с учетом отступов.
        Функция нужна для внутрених нужд
        :param start: начало диапозона
        :param end: конец диапозона
        :return:
        """

        delta = end - start

        return np.linspace(start - delta * self.left_offset, end + delta * self.right_offset, self.points)

    def gen_x_axis(self, start, end):
        """
        Генерирует набор точек по оси абсцисс в заданом диапозоне
        :param start: начало диапозона
        :param end: конец диапозона
        :return:
        """
        return np.linspace(start, end, self.points)

    def _prepare_before_approximation(self, x, y, xerr, yerr):
        if not isinstance(xerr, np.ndarray):
            xerr = np.ones_like(y) * xerr

        if not isinstance(yerr, np.ndarray):
            yerr = np.ones_like(y) * yerr

        self._x = np.array(x)
        self._y = np.array(y)
        self._xerr = xerr
        self._yerr = yerr

        return self._x, self._y, self._xerr, self._yerr

    def approximate(self, x, y, xerr=0, yerr=0):
        """
        Функция апроксимации

        :param x: набор параметров оси x
        :param y: набор параметров оси y
        :param xerr: погрешность параметров оси x
        :param yerr: погрешность параметров оси y

        :return: набор точек на кривой апроксимации
        """
        self._x = np.array(x)
        self._y = np.array(y)
        self._xerr = xerr
        self._yerr = yerr

        return x, y

    def get_function(self):
        """Возвращает полученную после аппроксимации функцию"""
        return self._function

    def label(self, xvar='x', yvar='y'):
        """
        Генерирует формулу для латеха

        :param xvar: буква перемонной по оси x
        :param yvar: буква перемонной по оси y

        :return: сгенерированная формулу
        """
        return f'[{self.__class__.__name__}] function with params: {self.meta}'

    def calc_hi_square(self):
        """
        Вычисляет chi^2
        """

        if not self._yerr.all():
            return None

        return np.sum(np.square((self._y - self.get_function()(self._x)) / self._yerr))

    def calc_quality_of_approximation(self):
        """
        Вычисляет chi^2 / (n - p)

        - n - количество точек в данных
        - p - количество степеней свободы (кол-во параметров в функции, которой аппроксимируем)
        """

        if not self._yerr.all():
            return None

        return self.calc_hi_square() / (len(self._x) - len(self.koefs))

    def is_approximation_good(self):
        """
        Делает оценочное суждение по поводу качества измерений

        quality_of_approximation = calc_quality_of_approximation = chi^2 / (n - p)

        - 0 < quality_of_approximation <= 0.5:

            Качество ваших измерений составляется quality_of_approximation <= 0.5.

            Это слишком мало, скорее всего это свидетельствуют о завышенных погрешностях

        - 0.5 < quality_of_approximation < 2:

            Качество ваших измерений составляется quality_of_approximation ~ 1.

            Это хороший результат. Ваша теоретическая модель хорошо сходится с экспериментом.

        - 2 <= quality_of_approximation:

            Качество ваших измерений составляется quality_of_approximation >= 2.

            Это слишком много. Это свидетельствуют либо о плохом соответствии
            теории и результатов измерений, либо о заниженных погрешностях.

        """
        if not self._yerr.all():
            return 'У ваших измерений по вертикальной оси нет погрешности, ' \
                   'поэтому невозможно пользоваться методом хи-квадрат'

        if len(self._x) - len(self.koefs) <= 0:
            return 'Слишком мало точек для аппроксимации, воспользуйтесь интерполяцией'

        quality_of_approximation = self.calc_quality_of_approximation()

        if 0 < quality_of_approximation <= 0.5:
            return f'Качество ваших измерений составляется {round(quality_of_approximation, 2)} <= 0.5.\n' \
                   f'Это слишком мало, скорее всего это свидетельствуют о завышенных погрешностях.'

        if 0.5 < quality_of_approximation < 2:
            return f'Качество ваших измерений составляется {round(quality_of_approximation, 2)} ~ 1.\n' \
                   f'Это хороший результат. Ваша теоретическая модель хорошо сходится с экспериментом.'

        if 2 <= quality_of_approximation:
            return f'Качество ваших измерений составляется {round(quality_of_approximation, 2)} >= 2.\n' \
                   f'Это слишком много. Это свидетельствуют либо о плохом соответствии \n' \
                   f'теории и результатов измерений, либо о заниженных погрешностях.'


class MultiLinearMixin:
    """
    Добавляет возможность получить функции,
    которая считает значения между двумя точками на прямой в результирующей аппроксимации

    Полезно для аппроксиматоров, которые только двигают точки, например Lowess
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._res_x = None
        self._res_y = None

    def get_function(self):

        def scalar_func(x):
            # При выходе за область определения возбуждаем исключение
            if x < np.min(self._x) or x > np.max(self._x):
                raise AttributeError(
                    f'Значение x={x} выходит за область определения [{np.min(self._x)}, {np.max(self._x)}]')

            # Если есть значение в точке
            idx = np.where(np.isclose(self._x, x))[0]

            if len(idx) != 0:
                return self._res_y[idx[0]]

            # Ищем значение на прямой между двумя ближайшими по оси x точками
            upper_bound = np.where(self._x > x)[0][0]
            # lower_bound = np.where(self._x < x)[0][-1]
            lower_bound = upper_bound - 1

            x1 = self._res_x[lower_bound]
            x2 = self._res_x[upper_bound]

            y1 = self._res_y[lower_bound]
            y2 = self._res_y[upper_bound]

            k = (y2 - y1) / (x2 - x1)
            return k * x + y1 - k * x1

        def vector_func(x):
            # При выходе за область определения возбуждаем исключение
            if np.any(x < np.min(self._x)) or np.any(x > np.max(self._x)):
                raise AttributeError(
                    f'Одна из координат x={x} выходит за область определения [{np.min(self._x)}, {np.max(self._x)}]')

            return np.array([scalar_func(_x_comp) for _x_comp in x])

        def inner(x):
            if isinstance(x, np.ndarray):
                return vector_func(x)

            else:
                return scalar_func(x)

        return inner


class Polynomial(Approximator):
    """
    Аппроксимация с помощью полинома. Использует numpy.polyfit
    """

    def __init__(self, deg=1, points=100, left_offset=5, right_offset=5):
        """
        :param deg: степень апроскимируещего полинома
        :param points: количество точек, которые будут на выходе
        :param left_offset: отступ от левой гриницы диапозона
        :param right_offset: отступ от правой гриницы диапозона
        """
        super(Polynomial, self).__init__(points, left_offset, right_offset)
        self.deg = deg

    def approximate(self, x, y, xerr=0, yerr=0):

        x, y, xerr, yerr = self._prepare_before_approximation(x, y, xerr, yerr)

        result = np.polyfit(x, y, w=1 / yerr, deg=self.deg, cov=True)

        popt = result[0]
        pcov = result[-1]

        self.meta = self.meta = {
            'popt': popt,
            'pcov': pcov
        }
        self.koefs = popt
        self.sigmas = np.sqrt(np.diag(pcov))

        self._function = np.poly1d(popt)

        poly = np.poly1d(popt)
        xs = self._gen_x_axis_with_offset(min(x), max(x))

        return xs, poly(xs)

    def label(self, xvar='x', yvar='y'):

        # если степень равна 0, то возвращаем эту константу
        if self.deg == 0:
            return f'${yvar} = {format_monoid(self.koefs[0], True)}$'

        # списисок моноидов
        monoids = []

        # форматируем коэффициент при каждой степени, кроме первой и нулевой
        for i in range(self.deg - 1):
            monoid = f'{format_monoid(self.koefs[i])}{xvar}^{{{self.deg - i}}}'
            monoids.append(monoid)

        # форматируем коэффициент при первой степени
        monoids.append(f'{format_monoid(self.koefs[self.deg - 1])}{xvar}')

        # форматируем коэффициент при нулевой степени
        monoids.append(f'{format_monoid(self.koefs[self.deg])}')

        # объединяем в один полином
        res = ''.join(monoids)

        # убираем плюс при максимальной степени
        # FIXME неоптимизированный костыль с копирование строк
        if self.koefs[0] >= 0:
            res = res[1:]

        return f"${yvar} = {res}$"


class Functional(Approximator):
    """
    Аппроксимация с помощью пользовательской функции. Использует scipy.optimize.curve_fit


    Функция должна иметь следующую сигнатуру:

    .. code:: python

        def function_for_fit(x, param1, param2, ..., paramn):
            ...

    x - переменная

    param1, param2, ..., paramn - параметры, которые будут искаться

    Например:

    .. code:: python

            def exp(x, a, b, c):
                return a * np.exp(b * x) + c

    У этой функции будут определяться параметры a, b, c
    """

    def __init__(self, function, points=100, left_offset=5, right_offset=5):
        """
        :param function: функция для аппроксимации
        :param points: количество точек, которые будут на выходе
        :param left_offset: отступ от левой границы диапазона
        :param right_offset: отступ от правой границы диапазона
        """
        super(Functional, self).__init__(points, left_offset, right_offset)
        self._function_for_fit = function

    def approximate(self, x, y, xerr=None, yerr=None):

        x, y, xerr, yerr = self._prepare_before_approximation(x, y, xerr, yerr)

        # пытаемся аппроксимировать. если у scipy не получается, то оно выбрасывает исключение RuntimeError
        try:
            popt, pcov = scipy.optimize.curve_fit(
                f=self._function_for_fit, xdata=x, ydata=y, sigma=yerr)
            perr = np.sqrt(np.diag(pcov))

            self.meta = {
                'popt': popt,
                'pcov': pcov
            }

            self.koefs = popt
            self.sigmas = perr

            def fff(*params):
                def inner(xx):
                    return self._function_for_fit(xx, *params)

                return inner

            self._function = fff(*self.koefs)

            xs = self._gen_x_axis_with_offset(min(x), max(x))
            ys = self._function_for_fit(xs, *self.koefs)
            return xs, ys

        except RuntimeError:
            # Если вызывается исключение, то возвращаем исходные данные
            warn(f"Точки плохо подходят под апроксимацию выбранной функцией {self._function.__name__}")
            return x, y

    def label(self, xvar='x', yvar='y'):
        try:
            return f'function with params: {[round_to_n(param, 3) for param in self.koefs]}'
        except IndexError:
            return f'Функция {self._function.__name__}, которая плохо подходит'



import matplotlib.pyplot as plt
import miptlabs
import numpy as np
import pandas as pd
from miptlabs import approximators
from miptlabs.latex import gen_from_dataframe, represent_value
from miptlabs.plotter import pretty_plot

pd, np, miptlabs, pretty_plot, gen_from_dataframe, plt, approximators, represent_value



import scipy.fftpack
import scipy.fftpack
import scipy.optimize
import scipy.optimize

from .base import Approximator, MultiLinearMixin

"""
Модуль содержаший базовые апроксиматоры
"""

from statsmodels.nonparametric.smoothers_lowess import lowess

import numpy as np
import scipy.fftpack
import scipy.optimize


class Lowess(MultiLinearMixin, Approximator):
    """
    Реализует алгоритм lowess.
    Предоставляет общий и гибкий подход для приближения двумерных данных.
    Подробнее_

    .. _Подробнее: http://www.machinelearning.ru/wiki/index.php?title=%D0%90%D0%BB%D0%B3%D0%BE%D1%80%D0%B8%D1%82%D0%BC_LOWESS

    Для некоторых наборов данных оченб хорошо апроксимирует кривую
    """

    def __init__(self, frac=0.35, points=100, left_offset=5, right_offset=5):
        """
        :param frac: Параметр f указывает, какая доля (fraction) данных используется в процедуре.
                Если f = 0.5, то только половина данных используется для оценки и влияет на результат,
                и тогда мы получим умеренное сглаживание. С другой стороны, если f = 0.8,
                то используются восемьдесят процентов данных, и сглаживание намного сильнее.
                Во всех случаях веса данных тем больше, чем они ближе к объекту t.
                Процедура оценки использует не метод наименьших квадратов,
                а более устойчивый ( робастный ) метод, который принимает меры против выбросов.
        :param points: количество точек, которые будут на выходе
        :param left_offset: отступ от левой гриницы диапозона
        :param right_offset: отступ от правой гриницы диапозона
        """
        super(Lowess, self).__init__(points, left_offset, right_offset)
        self.frac = frac

    def approximate(self, x, y, xerr=0, yerr=0):
        x, y, xerr, yerr = self._prepare_before_approximation(x, y, xerr, yerr)

        # Нечто
        result = lowess(y, x, frac=self.frac)

        self._res_x = result[:, 0]
        self._res_y = result[:, 1]

        return self._res_x, self._res_y


class Fourier(MultiLinearMixin, Approximator):
    """
    # WIP #

    Апроксимация функции с помощью преобразования фурье
    """

    def approximate(self, x, y, xerr=0, yerr=0):
        x, y, xerr, yerr = self._prepare_before_approximation(x, y, xerr, yerr)

        # Fourier
        x = np.array(x)
        y = np.array(y)
        w = scipy.fftpack.rfft(y)
        # f = scipy.fftpack.rfftfreq(10000, x[1] - x[0])
        spectrum = w ** 2
        cutoff_idx = spectrum < (spectrum.max() / 20)
        w2 = w.copy()
        w2[cutoff_idx] = 0
        y = scipy.fftpack.irfft(w2)

        self._res_x = x
        self._res_y = y

        return x, y



import numpy as np
import scipy.fftpack
import scipy.optimize

from . import functions
from .base import Polynomial, Functional, Approximator, MultiLinearMixin
from ..utils import round_to_n, format_monoid


class Linear(Polynomial):
    """
    Аппроксимация с помощью прямой y=kx+b. Использует numpy.polyfit
    """

    def __init__(self, points=100, left_offset=5, right_offset=5, no_bias=False):
        """
        :param points: количество точек, которые будут на выходе
        :param left_offset: отступ  от правой границы диапазона
        :param right_offset: отступ, то свободного коэффициента нет, иначе есть
        :param no_bias: Если *True* от левой границы диапазона
        """
        super(Linear, self).__init__(1, points, left_offset, right_offset)
        self.no_bias = no_bias

        # Данные
        self.__k = None
        self.__b = None

    def approximate(self, x, y, xerr=0, yerr=0):

        x, y, xerr, yerr = self._prepare_before_approximation(x, y, xerr, yerr)

        if not self.no_bias:
            return super(Linear, self).approximate(x, y, xerr, yerr)

        else:
            # fixme: убрать дублирование кода
            popt, pcov = scipy.optimize.curve_fit(
                f=functions.line_for_fit, xdata=x, ydata=y, sigma=yerr)
            perr = np.sqrt(np.diag(pcov))

            self.meta = {
                'popt': popt,
                'pcov': pcov
            }

            self.koefs = popt
            self.sigmas = perr

            self._function = functions.line(popt[0])

            xs = self._gen_x_axis_with_offset(min(x), max(x))
            ys = self._function(xs)
            return xs, ys

    def label(self, xvar='x', yvar='y'):
        res = f'{format_monoid(self.koefs[0])}{xvar}'

        if not self.no_bias:
            res += format_monoid(self.koefs[1])

        # убираем плюс при максимальной степени
        # FIXME неоптимизированный костыль с копирование строк
        if self.koefs[0] >= 0:
            res = res[1:]

        return f"${yvar} = {res}$"

    def _brac_x(self):
        return self._x.mean()

    def _brac_y(self):
        return self._y.mean()

    def _brac_x2(self):
        return (self._x * self._x).mean()

    def _brac_y2(self):
        return (self._y * self._y).mean()

    def _brac_xy(self):
        return (self._x * self._y).mean()

    def _d_xy(self):
        return (self._x - self._x.mean()).mean() * (self._y - self._y.mean()).mean()

    def _d_xx(self):
        return np.square(self._x - self._x.mean()).mean()

    def _d_yy(self):
        return np.square(self._y - self._y.mean()).mean()

    def _k(self):
        self.__k = (self._brac_xy() - self._brac_x() * self._brac_y()) / (self._brac_x2() - self._brac_x() ** 2)
        return self.__k

    def _b(self):
        if self.__k is None:
            self._k()

        self.__b = self._brac_y() - self.__k * self._brac_x()
        return self.__b

    def _sigma_k(self):
        if self.__k is None:
            self._k()

        if len(self._x) == 2:
            return 0

        return np.sqrt(np.abs((self._d_yy() / self._d_xx() - self._k() ** 2) / (len(self._x) - 2)))

    def _sigma_b(self):
        if self.__b is None:
            self._b()

        return self._sigma_k() * np.sqrt(self._brac_x2())


class Exponential(Functional):
    """
    Экспоненциальный апроксиматор
    y = a*exp(bx)+c
    """

    def __init__(self, points=100, left_offset=5, right_offset=5):
        """
        :param points: количество точек, которые будут на выходе
        :param left_offset: отступ  от правой границы диапазона
        :param right_offset: отступ, то свободного коэффициента нет, иначе есть
        """
        super(Exponential, self).__init__(
            function=functions.exp_for_fit,
            points=points,
            left_offset=left_offset,
            right_offset=right_offset
        )

    def label(self, xvar='x', yvar='y'):
        try:
            return f'${yvar} = {format_monoid(round_to_n(self.koefs[0], 3), True)}' \
                   f'e^{{{round_to_n(self.koefs[1], 3)}{xvar}}} ' \
                   f'{format_monoid(round_to_n(self.koefs[2], 3))}$'
        except TypeError:
            return 'Экспонента, которая не смогла'


class Logarithmic(Functional):
    """
    Логарифмический апроксиматор
    y = a * ln(bx + c) + d
    """

    def __init__(self, points=100, left_offset=5, right_offset=5):
        """
        :param points: количество точек, которые будут на выходе
        :param left_offset: отступ  от правой границы диапазона
        :param right_offset: отступ, то свободного коэффициента нет, иначе есть
        """
        super(Logarithmic, self).__init__(
            function=functions.log_for_fit,
            points=points,
            left_offset=left_offset,
            right_offset=right_offset
        )

    def label(self, xvar='x', yvar='y'):

        try:

            return f'${format_monoid(round_to_n(self.koefs[0], 3), True)}\ln{{(' \
                   f'{format_monoid(round_to_n(self.koefs[1], 3), True)}x ' \
                   f'{format_monoid(round_to_n(self.koefs[2], 3))})}} ' \
                   f'{format_monoid(round_to_n(self.koefs[3], 3))}$'

        except IndexError:
            return 'Логарифм, который не смог'


class TANH(Functional):
    """
    Логарифмический апроксиматор
    y = a * ln(bx + c) + d
    """

    def __init__(self, points=100, left_offset=5, right_offset=5):
        """
        :param points: количество точек, которые будут на выходе
        :param left_offset: отступ  от правой границы диапазона
        :param right_offset: отступ, то свободного коэффициента нет, иначе есть
        """
        super(TANH, self).__init__(
            function=functions.log_for_fit,
            points=points,
            left_offset=left_offset,
            right_offset=right_offset
        )

    def label(self, xvar='x', yvar='y'):

        try:

            return f'${format_monoid(round_to_n(self.koefs[0], 3), True)}\ln{{(' \
                   f'{format_monoid(round_to_n(self.koefs[1], 3), True)}x ' \
                   f'{format_monoid(round_to_n(self.koefs[2], 3))})}} ' \
                   f'{format_monoid(round_to_n(self.koefs[3], 3))}$'

        except IndexError:
            return 'Логарифм, который не смог'


class StupidApproximator(MultiLinearMixin, Approximator):
    """
    Аппроксиматор, который просто соединяет соседние точки прямой
    """



import matplotlib.pyplot as plt
from matplotlib.figure import Figure


class NothingDrawError(Exception):
    """
    Исключение, возбуждаемое, когда невозможно ничего нарисовать на графике.
    """

    def __init__(self, text):
        self.txt = text


def _make_tuples_from_filler(*tuples, length=1, filler=None):
    """
    Делает из filler кортеж длиной length, если передан не None, то оставляет его, как есть
    """
    result = []
    for tuple_ in tuples:
        if tuple_ is None:
            result.append(tuple([filler] * length))

        else:
            result.append(tuple_)

    return result


def _set_ticks_fontsize(axes, xfontsize, yfontsize):
    """
    Устанавливает размер шрифта подписям по осям
    :param axes: объект графика
    :param xfontsize: размер шрифта по оси x
    :param yfontsize: размер шрифта по оси y
    :return: None
    """
    for item in axes.get_xticklabels():
        item.set_fontsize(xfontsize)

    for item in axes.get_yticklabels():
        item.set_fontsize(yfontsize)


def _get_default_params():
    """
    Генерирует значения размеров шрифта по умолчанию
    :return:
    """
    params = dict()

    params['figsize'] = (10, 8)
    params['dpi'] = 100

    params['legend_loc'] = 'best'

    params['xticks_fontsize'] = 16
    params['yticks_fontsize'] = 16
    params['xlabel_fontsize'] = 22
    params['ylabel_fontsize'] = 22
    params['legend_fontsize'] = 20
    params['title_fontsize'] = 26

    return params


def _update_params(params, kwargs):
    """
    Обновляет значения параметров построения графика с учётом переданных дополнитеьных параметров
    :param params:
    :param kwargs:
    :return:
    """

    targets = ('figsize', 'dpi', 'legend_loc',
               'xticks_fontsize', 'yticks_fontsize',
               'xlabel_fontsize', 'ylabel_fontsize',
               'legend_fontsize', 'title_fontsize',
               'LEGEND_FROM_APPROXIMATOR', 'xvar', 'yvar')

    for target in targets:
        if target in kwargs:
            params[target] = kwargs.pop(target)

    if kwargs:
        print(f'Лишние аргументы', kwargs)


def _get_default_color(i=0):
    """
    :return: первый цвет в цикле цветов matplotlib, точнее тускло-синий
    """
    return plt.rcParams['axes.prop_cycle'].by_key()['color'][i % 10]


def _enable_minor_ticks(axes):
    """
    Включает побочную сетку на графике
    :param axes:
    :return:
    """
    axes.minorticks_on()
    axes.grid(which='major', color='k', linewidth=0.8)
    axes.grid(which='minor', color='k', linestyle=':')


def _set_font_params(axes, params, xlabel, ylabel, title):
    # Устанавливает размер шрифта для подписей по осям
    _set_ticks_fontsize(axes, params['xticks_fontsize'], params['yticks_fontsize'])

    # Устанавливает размер шрифта для разных элементов
    axes.set_xlabel(xlabel, fontsize=params['xlabel_fontsize'])
    axes.set_ylabel(ylabel, fontsize=params['ylabel_fontsize'])
    axes.set_title(title, fontsize=params['title_fontsize'])


def _get_axes(axes, params, minot_ticks):
    """
    Создаёт новые объект графика и фигуры, если axes=None
    Возвращает текущие оси, если они не None
    """
    if axes is None:
        figure, axes = plt.subplots(figsize=params['figsize'], dpi=params['dpi'])
        if minot_ticks:
            _enable_minor_ticks(axes)

    return axes


def _prepare(axes, xlabel, ylabel, title, minor_ticks, points, line, xerr, yerr, approximator, kwargs):
    _check_correct_input(points, line, xerr, yerr, approximator)

    # Инициализирует параметры
    params = _get_default_params()

    # Обновляет параметры
    _update_params(params, kwargs)

    # Создаёт новые объект графика и фигуры, если не переданы существующие
    axes = _get_axes(axes, params, minor_ticks)

    # ----------------Шрифты----------------
    _set_font_params(axes, params, xlabel, ylabel, title)

    return axes, params


def _check_correct_input(points, line, xerr, yerr, approximator):
    """
    Проверяет, может ли что-то нарисоваться.
    Если нет, то выбрасывает исключение NothingDrawError
    :param points:
    :param line:
    :param xerr:
    :param yerr:
    :return:
    """
    if not points and not line and not xerr and not yerr and not approximator:
        raise NothingDrawError("Хотя бы один из параметров points, line, xerr, yerr должен быть True,"
                               " чтобы что-то нарисовалось")


def _draw(axes, x, y, xerr, yerr, color, legend, points, line, approximator, params):
    """
    Сама функция рисования
    """
    # Устанавливает стандартный цвет, если не передан другой
    if color is None:
        color = _get_default_color()

    # Нужно для костыля, чтобы легенда отрисовыволась лишь один раз
    label = legend

    if approximator:
        x_, y_ = approximator.approximate(x, y, xerr, yerr)

        if params.get("LEGEND_FROM_APPROXIMATOR"):
            xvar = 'x'
            yvar = 'y'

            if 'xvar' in params:
                xvar = params['xvar']
            if 'yvar' in params:
                yvar = params['yvar']

            app_label = approximator.label(xvar, yvar)
        else:
            app_label = None

        axes.plot(x_, y_, c=color, label=app_label)

    # Рисует точки, если нужно
    if points:
        axes.scatter(x, y, c=color, label=label)
        label = None

    # Рисует кресты погрешностей, если нужно
    if xerr is not None or yerr is not None:
        axes.errorbar(x, y, fmt='none', xerr=xerr, yerr=yerr, c=color, label=label)
        label = None

    # Соединяет точки ломаной, если нужно
    if line:
        axes.plot(x, y, c=color, label=label)
        label = None

    return axes


def pretty_plot(x, y, xerr=0, yerr=0,
                xlabel=None, ylabel=None, title=None, legend=None,
                minor_ticks=True, color=None, points=True, line=False,
                axes=None, approximator=None, **kwargs):
    """
    Рисует график, с требованиями лабников.
    По умолчания не соединяет точки

    Без доп настроек можно рассчитывать на это

    .. figure:: _static/images/base.png
        :scale: 50 %
        :align: center
        :alt: Простой пример

    TODO: пока не умеет выносить степень. в будущем это появится.

    TODO: может поменяться интерфейс

    **Для отображения можно использовать:**

    - matplotlib.pyplot.show()
    - pretty_plot(...).figure.show()
    - show() (определён ниже)

    **Для сохранения графика в картинку можно использовать:**

    - matplotlib.pyplot.savefig(filename)
    - pretty_plot(...).figure.savefig(filename)
    - savefig(fig, filename) (определён ниже)

    :param x: координаты по оси x
    :param y: координаты по оси y
    :param xerr: погрешности по оси x.
        Либо одно число (применится к все точкам), либо список (применится к соответсвующей точке).
    :param yerr: погрешности по оси y.
        Либо одно число (применится к все точкам), либо список (применится к соответсвующей точке).
    :param xlabel: подпись по оси x
    :param ylabel: подпись по оси y
    :param title: название графика
    :param legend: легенда
    :param minor_ticks: нужна ли вспомогательная сетка
    :param color: цвет графика. Если None, то будет синий
    :param points: нужно ли рисовать точки
    :param line: нужно ли соединить ломаной все точки
    :param axes: объект графика. Можно передать, чтобы дорисовать всё на существующем графике
    :param approximator: аппроксиматор для точек, должен быть экземпляром класса Approximator
    :param kwargs: дополнительные аргументы

    :Дополнительные параметры:

    * *figsize* (``tuple[int]``) --
      размер графика в дюймах, по умолчанию 10 на 8
    * *dpi* (``int``) --
      количество пикселей на дюйм, по умолчанию 100
    * *legend_loc* (``str``) --
      положение легенды (см. документацию matplotlib), по умолчанию best
    * *xticks_fontsize* (``int``) --
      размер шрифта подписей оси x, по умолчанию 16
    * *yticks_fontsize* (``int``) --
      размер шрифта подписей оси y, по умолчанию 16
    * *xlabel_fontsize* (``int``) --
      размер шрифта обозначения оси x, по умолчанию 22
    * *ylabel_fontsize* (``int``) --
      размер шрифта обозначения оси y, по умолчанию 22
    * *title_fontsize* (``int``) --
      размер шрифта заголовка, по умолчанию 26
    * *legend_fontsize* (``int``) --
      размер шрифта легенды, по умолчанию 22
    * *LEGEND_FROM_APPROXIMATOR* (``int``) --
      если передано True, то рисует легенду ещё из аппроксиматора
    * *xvar* (``str``) --
      величина по оси x, если легенда рисуется из аппроксиматора
    * *yvar* (``str``) --
      величина по оси y, если легенда рисуется из аппроксиматора

    :return: экземпляр класса matplotlib.Axes - объект только что нарисованного графика
    """

    # ----------------Инициализация----------------
    axes, params = _prepare(
        axes=axes,
        xlabel=xlabel,
        ylabel=ylabel,
        title=title,
        minor_ticks=minor_ticks,
        points=points,
        line=line,
        xerr=xerr,
        yerr=yerr,
        approximator=approximator,
        kwargs=kwargs
    )

    # ----------------Рисование----------------
    _draw(axes, x, y, xerr, yerr, color, legend, points, line, approximator, params)

    # Рисует легенду, если нужно
    if legend or params.get("LEGEND_FROM_APPROXIMATOR"):
        axes.legend(loc=params['legend_loc'], fontsize=params['legend_fontsize'])

    return axes


def pretty_plot_many(xs, ys, xerrs=None, yerrs=None,
                     xlabel=None, ylabel=None, title=None, legends=None,
                     minor_ticks=True, colors=None, points=True, line=False,
                     axes=None, approximator=None, **kwargs):
    r"""
    Рисует график так же, как и pretty_plot,
    только вместо x, y передаётся два списка с наборами координат,
    что позволяет сразу отрисовать несколько графиков

    :param xs: наборы координат по оси x
    :param ys: наборы координат по оси y
    :param xerrs: наборы погрешностей по оси x.
        Либо одно число (применится к все точкам), либо список (применится к соответсвующей точке).
    :param yerrs: наборы погрешностей по оси y.
        Либо одно число (применится к все точкам), либо список (применится к соответсвующей точке).
    :param xlabel: наборы подписей по оси x
    :param ylabel: наборы подписей по оси y
    :param title: название графика
    :param legends: легенды
    :param minor_ticks: нужна ли впсомогательная сетка
    :param colors: наборы цветов графика. Если None, то будет синий
    :param points: нужно ли рисвоать точки
    :param line: нужно ли соединить ломаной все точки
    :param axes: объект графика. Можно передать, чтобы дорисовать всё на существующем графике
    :param approximator: аппроксиматор для точек, должен быть экземпляром класса Approximator
    :param kwargs: дополнительные агрменты:

    :Дополнительные параметры:

    * *figsize* (``tuple[int]``) --
      размер графика в дюймах, по умолчанию 10 на 8
    * *dpi* (``int``) --
      количество пикселей на дюйм, по умолчанию 100
    * *legend_loc* (``str``) --
      положение легенды (см. документацию matplotlib), по умолчанию best
    * *xticks_fontsize* (``int``) --
      размер шрифта подписей оси x, по умолчанию 16
    * *yticks_fontsize* (``int``) --
      размер шрифта подписей оси y, по умолчанию 16
    * *xlabel_fontsize* (``int``) --
      размер шрифта обозначения оси x, по умолчанию 22
    * *ylabel_fontsize* (``int``) --
      размер шрифта обозначения оси y, по умолчанию 22
    * *title_fontsize* (``int``) --
      размер шрифта заголовка, по умолчанию 26
    * *legend_fontsize* (``int``) --
      размер шрифта легенды, по умолчанию 22
    * *LEGEND_FROM_APPROXIMATOR* (``bool``) --
      если передано True, то рисует легенду ещё из аппроксиматора
    * *xvar* (``str``) --
      величина по оси x, если легенда рисуется из аппроксиматора
    * *yvar* (``str``) --
      величина по оси y, если легенда рисуется из аппроксиматора

    :return: экземпляр класса matplotlib.Axes - объект только что нарисованного графика
    """

    # ----------------Инициализация----------------
    axes, params = _prepare(
        axes=axes,
        xlabel=xlabel,
        ylabel=ylabel,
        title=title,
        minor_ticks=minor_ticks,
        points=points,
        line=line,
        xerr=xerrs,
        yerr=yerrs,
        approximator=approximator,
        kwargs=kwargs
    )

    # Создаём кортежы из None или нулей длины = количеству графиков
    legends, colors = _make_tuples_from_filler(legends, colors)
    xerrs, yerrs = _make_tuples_from_filler(xerrs, yerrs, filler=0)

    # ----------------Рисование----------------
    for num, (x, y, xerr, yerr, legend, color) in enumerate(zip(xs, ys, xerrs, yerrs, legends, colors)):
        _draw(axes, x, y, xerr, yerr, color, legend, points, line, approximator, params)

    # Рисует легенду, если нужно
    if legends or params.get("LEGEND_FROM_APPROXIMATOR"):
        axes.legend(loc=params['legend_loc'], fontsize=params['legend_fontsize'])

    return axes


def show(*args, **kwargs):
    """
    Рисует все сгенерированные графики. По-сути обертка над matplotlib.pyplot.show

    Подробнее в `документации matplotlib`_

    .. _`документации matplotlib`: https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.show.html
    """
    return plt.show(*args, **kwargs)


def savefig(obj, filename, *, transparent=None, **kwargs):
    """
    Сохраняет фигуру. По-сути обертка над встроенной функцие сохранения

    :param obj: экземпляр класса Axes или Figure из matplotlib. Его же возвращают функции pretty_plot и pretty_plot_many
    :param filename: имя файла, в который нужно сохранить. Если передан без расширения, то сохраниться в png.
    :param transparent: If *True*, the axes patches will all be transparent; the
            figure patch will also be transparent unless facecolor
            and/or edgecolor are specified via kwargs.
            This is useful, for example, for displaying
            a plot on top of a colored background on a web page.  The
            transparency of these patches will be restored to their
            original values upon exit of this function.
    :param kwargs: дополнительные параметры. Смотри `документацию matplotlib`_

    .. _`документацию matplotlib`: https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.savefig.html
    """
    figure = obj if isinstance(obj, Figure) else obj.figure

    figure.savefig(filename, transparent=transparent, **kwargs)



import os
import sys

import setuptools

sys.path.insert(0, os.path.abspath('.'))
from src.miptlabs import __version__

with open("readme.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.read().strip().split('\n')

setuptools.setup(
    name='miptlabs',
    version=__version__,
    author="Dmitry",
    description="The package will help MIPT students to draw graphs easier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dimon58/miptlabs",
    project_urls={
        "Bug Tracker": "https://github.com/dimon58/miptlabs/issues",
        "Documentation": "https://miptlabs.readthedocs.io/",
    },
    data_files=[
        'readme.md',
        'requirements.txt'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=requirements
)
